use std::hint::cold_path;

use crate::{
    job::{JobState, StackJob},
    latch::{AsCoreLatch, LatchRef, WakeLatch},
    workerthread::WorkerThread,
};

impl WorkerThread {
    #[inline]
    fn join_seq<A, B, RA, RB>(&self, a: A, b: B) -> (RA, RB)
    where
        RA: Send,
        RB: Send,
        A: FnOnce() -> RA + Send,
        B: FnOnce() -> RB + Send,
    {
        let rb = b();
        let ra = a();

        (ra, rb)
    }

    /// This function must be called from a worker thread.
    #[inline]
    pub(crate) fn join_heartbeat_every<A, B, RA, RB, const TIMES: usize>(
        &self,
        a: A,
        b: B,
    ) -> (RA, RB)
    where
        RA: Send,
        RB: Send,
        A: FnOnce() -> RA + Send,
        B: FnOnce() -> RB + Send,
    {
        // SAFETY: each worker is only ever used by one thread, so this is safe.
        let count = self.join_count.get();
        self.join_count.set(count.wrapping_add(1) % TIMES as u8);

        // TODO: add counter to job queue, check for low job count to decide whether to use heartbeat or seq.
        // see: chili

        // SAFETY: this function runs in a worker thread, so we can access the queue safely.
        if count == 0 || unsafe { self.queue.as_ref_unchecked().len() } < 3 {
            cold_path();
            self.join_heartbeat(a, b)
        } else {
            self.join_seq(a, b)
        }
    }

    /// This function must be called from a worker thread.
    #[inline]
    fn join_heartbeat<A, B, RA, RB>(&self, a: A, b: B) -> (RA, RB)
    where
        RA: Send,
        RB: Send,
        A: FnOnce() -> RA + Send,
        B: FnOnce() -> RB + Send,
    {
        use std::panic::{AssertUnwindSafe, catch_unwind, resume_unwind};

        let latch = WakeLatch::new(self.context.clone(), self.index);
        let a = StackJob::new(
            move || {
                // TODO: bench whether tick'ing here is good.
                // turns out this actually costs a lot of time, likely because of the thread local check.
                // WorkerThread::current_ref()
                //     .expect("stackjob is run in workerthread.")
                //     .tick();

                a()
            },
            LatchRef::new(&latch),
        );

        let job = a.as_job();
        self.push_front(&job);

        let rb = match catch_unwind(AssertUnwindSafe(|| b())) {
            Ok(val) => val,
            Err(payload) => {
                cold_path();
                // if b panicked, we need to wait for a to finish
                self.wait_until_latch(&latch);
                resume_unwind(payload);
            }
        };

        let ra = if job.state() == JobState::Empty as u8 {
            unsafe {
                job.unlink();
            }

            // a is allowed to panic here, because we already finished b.
            unsafe { a.unwrap()() }
        } else {
            match self.wait_until_job::<RA>(unsafe { job.transmute_ref() }, latch.as_core_latch()) {
                Some(t) => t.into_result(), // propagate panic here
                // the job was shared, but not yet stolen, so we get to run the
                // job inline
                None => unsafe { a.unwrap()() },
            }
        };

        drop(a);
        (ra, rb)
    }
}
