use core::{
    marker::PhantomData,
    sync::atomic::{AtomicUsize, Ordering},
};
use std::sync::{Arc, atomic::AtomicU8};

use parking_lot::{Condvar, Mutex};

use crate::context::Context;

pub trait Latch {
    unsafe fn set_raw(this: *const Self);
}

pub trait Probe {
    fn probe(&self) -> bool;
}

pub type CoreLatch = AtomicLatch;
pub trait AsCoreLatch {
    fn as_core_latch(&self) -> &CoreLatch;
}

#[derive(Debug)]
pub struct AtomicLatch {
    inner: AtomicU8,
}

impl AtomicLatch {
    pub const UNSET: u8 = 0;
    pub const SET: u8 = 1;
    pub const SLEEPING: u8 = 2;

    #[inline]
    pub const fn new() -> Self {
        Self {
            inner: AtomicU8::new(Self::UNSET),
        }
    }
    #[inline]
    pub fn reset(&self) {
        self.inner.store(Self::UNSET, Ordering::Release);
    }

    pub fn get(&self) -> u8 {
        self.inner.load(Ordering::Acquire)
    }

    /// returns true if the latch was previously sleeping.
    #[inline]
    pub unsafe fn set(this: *const Self) -> bool {
        unsafe {
            let old = (*this).inner.swap(Self::SET, Ordering::Release);
            old == Self::SLEEPING
        }
    }
}

impl Latch for AtomicLatch {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            Self::set(this);
        }
    }
}

impl Probe for AtomicLatch {
    #[inline]
    fn probe(&self) -> bool {
        self.inner.load(Ordering::Acquire) == Self::SET
    }
}
impl AsCoreLatch for AtomicLatch {
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        self
    }
}

pub struct LatchRef<'a, L: Latch> {
    inner: *const L,
    _marker: PhantomData<&'a L>,
}

impl<'a, L: Latch> LatchRef<'a, L> {
    #[inline]
    pub const fn new(latch: &'a L) -> Self {
        Self {
            inner: latch,
            _marker: PhantomData,
        }
    }
}

impl<'a, L: Latch> Latch for LatchRef<'a, L> {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            let this = &*this;
            Latch::set_raw(this.inner);
        }
    }
}

impl<'a, L: Latch + Probe> Probe for LatchRef<'a, L> {
    #[inline]
    fn probe(&self) -> bool {
        unsafe {
            let this = &*self.inner;
            Probe::probe(this)
        }
    }
}

impl<'a, L> AsCoreLatch for LatchRef<'a, L>
where
    L: Latch + AsCoreLatch,
{
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        unsafe {
            let this = &*self.inner;
            this.as_core_latch()
        }
    }
}

pub struct NopLatch;

impl Latch for NopLatch {
    #[inline]
    unsafe fn set_raw(_this: *const Self) {
        // do nothing
    }
}

impl Probe for NopLatch {
    #[inline]
    fn probe(&self) -> bool {
        false // always returns false
    }
}

pub struct ThreadWakeLatch {
    waker: Mutex<Option<std::thread::Thread>>,
}

impl ThreadWakeLatch {
    #[inline]
    pub const fn new() -> Self {
        Self {
            waker: Mutex::new(None),
        }
    }

    #[inline]
    pub fn reset(&self) {
        let mut waker = self.waker.lock();
        *waker = None;
    }

    #[inline]
    pub fn set_waker(&self, thread: std::thread::Thread) {
        let mut waker = self.waker.lock();
        *waker = Some(thread);
    }

    pub unsafe fn wait(&self) {
        assert!(
            self.waker.lock().replace(std::thread::current()).is_none(),
            "ThreadWakeLatch can only be waited once per thread"
        );

        std::thread::park();
    }
}

impl Latch for ThreadWakeLatch {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            if let Some(thread) = (&*this).waker.lock().take() {
                thread.unpark();
            }
        }
    }
}

impl Probe for ThreadWakeLatch {
    #[inline]
    fn probe(&self) -> bool {
        self.waker.lock().is_some()
    }
}

pub struct CountLatch<L: Latch> {
    count: AtomicUsize,
    inner: L,
}

impl<L: Latch> CountLatch<L> {
    #[inline]
    pub const fn new(inner: L) -> Self {
        Self {
            count: AtomicUsize::new(0),
            inner,
        }
    }

    pub fn count(&self) -> usize {
        self.count.load(Ordering::Relaxed)
    }

    pub fn inner(&self) -> &L {
        &self.inner
    }

    #[inline]
    pub fn increment(&self) {
        self.count.fetch_add(1, Ordering::Release);
    }

    #[inline]
    pub fn decrement(&self) {
        if self.count.fetch_sub(1, Ordering::Release) == 1 {
            unsafe {
                Latch::set_raw(&self.inner);
            }
        }
    }
}

impl<L: Latch> Latch for CountLatch<L> {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            let this = &*this;
            this.decrement();
        }
    }
}

impl<L: Latch + Probe> Probe for CountLatch<L> {
    #[inline]
    fn probe(&self) -> bool {
        self.inner.probe()
    }
}

impl<L: Latch + AsCoreLatch> AsCoreLatch for CountLatch<L> {
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        self.inner.as_core_latch()
    }
}

pub struct MutexLatch {
    inner: Mutex<bool>,
    condvar: Condvar,
}

impl MutexLatch {
    #[inline]
    pub const fn new() -> Self {
        Self {
            inner: Mutex::new(false),
            condvar: Condvar::new(),
        }
    }

    #[inline]
    pub fn reset(&self) {
        let mut guard = self.inner.lock();
        *guard = false;
    }

    pub fn wait(&self) {
        let mut guard = self.inner.lock();
        while !*guard {
            self.condvar.wait(&mut guard);
        }
    }

    pub fn set(&self) {
        unsafe {
            Latch::set_raw(self);
        }
    }

    pub fn wait_and_reset(&self) {
        let mut guard = self.inner.lock();
        while !*guard {
            self.condvar.wait(&mut guard);
        }
        *guard = false;
    }
}

impl Latch for MutexLatch {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            *(&*this).inner.lock() = true;
            (&*this).condvar.notify_all();
        }
    }
}

impl Probe for MutexLatch {
    #[inline]
    fn probe(&self) -> bool {
        *self.inner.lock()
    }
}

pub struct WakeLatch {
    inner: AtomicLatch,
    context: Arc<Context>,
    worker_index: AtomicUsize,
}

impl WakeLatch {
    pub fn new(context: Arc<Context>, worker_index: usize) -> Self {
        Self {
            inner: AtomicLatch::new(),
            context,
            worker_index: AtomicUsize::new(worker_index),
        }
    }

    pub(crate) fn set_worker_index(&self, worker_index: usize) {
        self.worker_index.store(worker_index, Ordering::Relaxed);
    }
}

impl Latch for WakeLatch {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            let ctx = (&*this).context.clone();
            let worker_index = (&*this).worker_index.load(Ordering::Relaxed);

            if CoreLatch::set(&(&*this).inner) {
                // If the latch was sleeping, wake the worker thread
                ctx.shared().heartbeats.get(&worker_index).and_then(|weak| {
                    weak.upgrade()
                        .map(|heartbeat| Latch::set_raw(&heartbeat.latch))
                });
            }
        }
    }
}

impl Probe for WakeLatch {
    #[inline]
    fn probe(&self) -> bool {
        self.inner.probe()
    }
}

impl AsCoreLatch for WakeLatch {
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        &self.inner
    }
}
