use std::{
    any::Any,
    marker::PhantomData,
    ptr::{self, NonNull},
    sync::{
        Arc,
        atomic::{AtomicPtr, Ordering},
    },
};

use async_task::Runnable;

use crate::{
    context::{Context, run_in_worker},
    job::{HeapJob, Job},
    latch::{AsCoreLatch, CountLatch, WakeLatch},
    util::{DropGuard, SendPtr},
    workerthread::WorkerThread,
};

pub struct Scope<'scope> {
    // latch to wait on before the scope finishes
    job_counter: CountLatch<WakeLatch>,
    // local threadpool
    context: Arc<Context>,
    // panic error
    panic: AtomicPtr<Box<dyn Any + Send + 'static>>,
    // variant lifetime
    _pd: PhantomData<fn(&'scope ())>,
}

pub fn scope<'scope, F, R>(f: F) -> R
where
    F: FnOnce(&Scope<'scope>) -> R + Send,
    R: Send,
{
    Scope::<'scope>::scope(f)
}

impl<'scope> Scope<'scope> {
    fn wait_for_jobs(&self, worker: &WorkerThread) {
        tracing::trace!("waiting for {} jobs to finish.", self.job_counter.count());
        tracing::trace!("thread id: {:?}, jobs: {:?}", worker.index, unsafe {
            worker.queue.as_ref_unchecked()
        });

        // set worker index in the job counter
        self.job_counter.inner().set_worker_index(worker.index);
        worker.wait_until_latch(self.job_counter.as_core_latch());
    }

    pub fn scope<F, R>(f: F) -> R
    where
        F: FnOnce(&Self) -> R + Send,
        R: Send,
    {
        run_in_worker(|worker| {
            // SAFETY: we call complete() after creating this scope, which
            // ensures that any jobs spawned from the scope exit before the
            // scope closes.
            let this = unsafe { Self::from_context(worker.context.clone()) };
            this.complete(worker, || f(&this))
        })
    }

    fn scope_with_context<F, R>(context: Arc<Context>, f: F) -> R
    where
        F: FnOnce(&Self) -> R + Send,
        R: Send,
    {
        context.run_in_worker(|worker| {
            // SAFETY: we call complete() after creating this scope, which
            // ensures that any jobs spawned from the scope exit before the
            // scope closes.
            let this = unsafe { Self::from_context(context.clone()) };
            this.complete(worker, || f(&this))
        })
    }

    /// should be called from within a worker thread.
    fn complete<F, R>(&self, worker: &WorkerThread, f: F) -> R
    where
        F: FnOnce() -> R + Send,
        R: Send,
    {
        use std::panic::{AssertUnwindSafe, catch_unwind};

        #[allow(dead_code)]
        fn make_job<F: FnOnce() -> T, T>(f: F) -> Job<T> {
            #[align(8)]
            unsafe fn harness<F: FnOnce() -> T, T>(this: *const (), job: *const Job<T>) {
                let f = unsafe { Box::from_raw(this.cast::<F>().cast_mut()) };

                let result = catch_unwind(AssertUnwindSafe(move || f()));

                let job = unsafe { Box::from_raw(job.cast_mut()) };
                job.complete(result);
            }

            Job::<T>::new(harness::<F, T>, unsafe {
                NonNull::new_unchecked(Box::into_raw(Box::new(f))).cast()
            })
        }

        let result = match catch_unwind(AssertUnwindSafe(|| f())) {
            Ok(val) => Some(val),
            Err(payload) => {
                self.panicked(payload);
                None
            }
        };

        self.wait_for_jobs(worker);
        self.maybe_propagate_panic();

        // SAFETY: if result panicked, we would have propagated the panic above.
        result.unwrap()
    }

    /// resumes the panic if one happened in this scope.
    fn maybe_propagate_panic(&self) {
        let err_ptr = self.panic.load(Ordering::Relaxed);
        if !err_ptr.is_null() {
            unsafe {
                let err = Box::from_raw(err_ptr);
                std::panic::resume_unwind(*err);
            }
        }
    }

    /// stores the first panic that happened in this scope.
    fn panicked(&self, err: Box<dyn Any + Send + 'static>) {
        self.panic.load(Ordering::Relaxed).is_null().then(|| {
            use core::mem::ManuallyDrop;
            let mut boxed = ManuallyDrop::new(Box::new(err));

            let err_ptr: *mut Box<dyn Any + Send + 'static> = &mut **boxed;
            if self
                .panic
                .compare_exchange(
                    ptr::null_mut(),
                    err_ptr,
                    Ordering::SeqCst,
                    Ordering::Relaxed,
                )
                .is_ok()
            {
                // we successfully set the panic, no need to drop
            } else {
                // drop the error, someone else already set it
                _ = ManuallyDrop::into_inner(boxed);
            }
        });
    }

    pub fn spawn<F>(&self, f: F)
    where
        F: FnOnce(&Scope<'scope>) + Send,
    {
        self.context.run_in_worker(|worker| {
            self.job_counter.increment();

            let this = SendPtr::new_const(self).unwrap();

            let job = Box::new(HeapJob::new(move || unsafe {
                _ = f(this.as_ref());
                this.as_ref().job_counter.decrement();
            }))
            .into_boxed_job();

            tracing::trace!("allocated heapjob");

            worker.push_front(job);

            tracing::trace!("leaked heapjob");
        });
    }

    pub fn spawn_future<T, F>(&self, future: F) -> async_task::Task<T>
    where
        F: Future<Output = T> + Send + 'scope,
        T: Send + 'scope,
    {
        self.context.run_in_worker(|worker| {
            self.job_counter.increment();

            let this = SendPtr::new_const(&self.job_counter).unwrap();

            let future = async move {
                let _guard = DropGuard::new(move || unsafe {
                    this.as_ref().decrement();
                });
                future.await
            };

            let schedule = move |runnable: Runnable| {
                #[align(8)]
                unsafe fn harness<T>(this: *const (), job: *const Job<T>) {
                    unsafe {
                        let runnable =
                            Runnable::<()>::from_raw(NonNull::new_unchecked(this.cast_mut()));
                        runnable.run();

                        // SAFETY: job was turned into raw
                        drop(Box::from_raw(job.cast_mut()));
                    }
                }

                let job = Box::new(Job::<T>::new(harness::<T>, runnable.into_raw()));

                // casting into Job<()> here
                worker.push_front(Box::into_raw(job) as _);
            };

            let (runnable, task) = unsafe { async_task::spawn_unchecked(future, schedule) };

            runnable.schedule();

            task
        })
    }

    #[allow(dead_code)]
    fn spawn_async<'a, T, Fut, Fn>(&'a self, f: Fn) -> async_task::Task<T>
    where
        Fn: FnOnce(&Scope) -> Fut + Send + 'static,
        Fut: Future<Output = T> + Send + 'static,
        T: Send + 'static,
    {
        let this = SendPtr::new_const(self).unwrap();
        let future = async move { f(unsafe { this.as_ref() }).await };

        self.spawn_future(future)
    }

    #[inline]
    pub fn join<A, B, RA, RB>(&self, a: A, b: B) -> (RA, RB)
    where
        RA: Send,
        RB: Send,
        A: FnOnce(&Self) -> RA + Send,
        B: FnOnce(&Self) -> RB + Send,
    {
        let worker = WorkerThread::current_ref().expect("join is run in workerthread.");
        let this = SendPtr::new_const(self).unwrap();

        worker.join_heartbeat_every::<_, _, _, _, 64>(
            {
                let this = this;
                move || a(unsafe { this.as_ref() })
            },
            {
                let this = this;
                move || b(unsafe { this.as_ref() })
            },
        )
    }

    unsafe fn from_context(ctx: Arc<Context>) -> Self {
        Self {
            context: ctx.clone(),
            job_counter: CountLatch::new(WakeLatch::new(ctx, 0)),
            panic: AtomicPtr::new(ptr::null_mut()),
            _pd: PhantomData,
        }
    }
}
