use std::{
    cell::{Cell, UnsafeCell},
    ptr::NonNull,
    sync::Arc,
    time::Duration,
};

use crossbeam_utils::CachePadded;
use parking_lot_core::SpinWait;

use crate::{
    context::{Context, Heartbeat},
    job::{Job, JobList, JobResult},
    latch::{AsCoreLatch, CoreLatch, Probe},
    util::DropGuard,
};

pub struct WorkerThread {
    pub(crate) context: Arc<Context>,
    pub(crate) index: usize,
    pub(crate) queue: UnsafeCell<JobList>,
    heartbeat: Arc<CachePadded<Heartbeat>>,
    pub(crate) join_count: Cell<u8>,
}

thread_local! {
    static WORKER: UnsafeCell<Option<NonNull<WorkerThread>>> = const { UnsafeCell::new(None) };
}

impl WorkerThread {
    pub fn new_in(context: Arc<Context>) -> Self {
        let (heartbeat, index) = context.shared().new_heartbeat();

        Self {
            context,
            index,
            queue: UnsafeCell::new(JobList::new()),
            heartbeat,
            join_count: Cell::new(0),
        }
    }

    fn new() -> Self {
        let context = Context::global_context().clone();
        Self::new_in(context)
    }
}

impl WorkerThread {
    pub fn run(self: Box<Self>) {
        let this = Box::into_raw(self);
        unsafe {
            Self::set_current(this);
        }

        let _guard = DropGuard::new(|| unsafe {
            // SAFETY: this is only called when the thread is exiting
            Self::unset_current();
            Self::drop_in_place(this);
        });

        tracing::trace!("WorkerThread::run: starting worker thread");

        unsafe {
            (&*this).run_inner();
        }

        tracing::trace!("WorkerThread::run: worker thread finished");
    }

    fn run_inner(&self) {
        let mut job = self.context.shared().pop_job();
        'outer: loop {
            let mut guard = loop {
                if let Some(job) = job {
                    self.execute(job);
                }

                let mut guard = self.context.shared();
                if guard.should_exit() {
                    // if the context is stopped, break out of the outer loop which
                    // will exit the thread.
                    break 'outer;
                }

                match guard.pop_job() {
                    Some(job) => {
                        tracing::trace!("worker: popping job: {:?}", job);
                        // found job, continue inner loop
                        continue;
                    }
                    None => {
                        tracing::trace!("worker: no job, waiting for shared job");
                        // no more jobs, break out of inner loop and wait for shared job
                        break guard;
                    }
                }
            };

            self.context.shared_job.wait(&mut guard);
            job = guard.pop_job();
        }
    }
}

impl WorkerThread {
    #[inline(always)]
    fn tick(&self) {
        if self.heartbeat.is_pending() {
            self.heartbeat_cold();
        }
    }

    #[inline]
    fn execute(&self, job: NonNull<Job>) {
        self.tick();
        Job::execute(job);
    }

    #[cold]
    fn heartbeat_cold(&self) {
        let mut guard = self.context.shared();

        if !guard.jobs.contains_key(&self.index) {
            if let Some(job) = self.pop_back() {
                tracing::trace!("heartbeat: sharing job: {:?}", job);
                unsafe {
                    job.as_ref().set_pending();
                }
                guard.jobs.insert(self.index, job);
                self.context.notify_shared_job();
            }
        }

        self.heartbeat.clear();
    }
}

impl WorkerThread {
    #[inline]
    pub fn pop_back(&self) -> Option<NonNull<Job>> {
        unsafe { self.queue.as_mut_unchecked().pop_back() }
    }

    #[inline]
    pub fn push_back(&self, job: *const Job) {
        unsafe { self.queue.as_mut_unchecked().push_back(job) }
    }

    #[inline]
    pub fn pop_front(&self) -> Option<NonNull<Job>> {
        unsafe { self.queue.as_mut_unchecked().pop_front() }
    }

    #[inline]
    pub fn push_front(&self, job: *const Job) {
        unsafe { self.queue.as_mut_unchecked().push_front(job) }
    }
}

impl WorkerThread {
    #[inline]
    pub fn current_ref<'a>() -> Option<&'a Self> {
        unsafe { (*WORKER.with(UnsafeCell::get)).map(|ptr| ptr.as_ref()) }
    }

    unsafe fn set_current(this: *const Self) {
        WORKER.with(|cell| {
            unsafe {
                // SAFETY: this cell is only ever accessed from the current thread
                assert!(
                    (&mut *cell.get())
                        .replace(NonNull::new_unchecked(
                            this as *const WorkerThread as *mut WorkerThread,
                        ))
                        .is_none()
                );
            }
        });
    }

    unsafe fn unset_current() {
        WORKER.with(|cell| {
            unsafe {
                // SAFETY: this cell is only ever accessed from the current thread
                (&mut *cell.get()).take();
            }
        });
    }

    unsafe fn drop_in_place(this: *mut Self) {
        unsafe {
            this.drop_in_place();
            drop(Box::from_raw(this));
        }
    }
}

pub struct HeartbeatThread {
    ctx: Arc<Context>,
}

impl HeartbeatThread {
    const HEARTBEAT_INTERVAL: Duration = Duration::from_micros(100);

    pub fn new(ctx: Arc<Context>) -> Self {
        Self { ctx }
    }

    pub fn run(self) {
        tracing::trace!("new heartbeat thread {:?}", std::thread::current());

        let mut i = 0;
        loop {
            let sleep_for = {
                let mut guard = self.ctx.shared();
                if guard.should_exit() {
                    break;
                }

                let mut n = 0;
                guard.heartbeats.retain(|_, b| {
                    b.upgrade()
                        .inspect(|heartbeat| {
                            if n == i {
                                if heartbeat.set_pending() {
                                    heartbeat.latch.set();
                                }
                            }
                            n += 1;
                        })
                        .is_some()
                });
                let num_heartbeats = guard.heartbeats.len();

                drop(guard);

                if i >= num_heartbeats {
                    i = 0;
                } else {
                    i += 1;
                }

                Self::HEARTBEAT_INTERVAL.checked_div(num_heartbeats as u32)
            };

            if let Some(duration) = sleep_for {
                std::thread::sleep(duration);
            }
        }
    }
}

impl WorkerThread {
    #[cold]
    fn wait_until_latch_cold(&self, latch: &CoreLatch) {
        // does this optimise?
        assert!(!latch.probe());

        'outer: while !latch.probe() {
            // take a shared job, if it exists
            if let Some(shared_job) = self.context.shared().jobs.remove(&self.index) {
                self.execute(shared_job);
            }

            // process local jobs before locking shared context
            while let Some(job) = self.pop_front() {
                unsafe {
                    job.as_ref().set_pending();
                }
                self.execute(job);
            }

            while !latch.probe() {
                let job = self.context.shared().pop_job();

                match job {
                    Some(job) => {
                        self.execute(job);

                        continue 'outer;
                    }
                    None => {
                        tracing::trace!("waiting for shared job, thread id: {:?}", self.index);

                        // TODO: wait on latch? if we have something that can
                        // signal being done, e.g. can be waited on instead of
                        // shared jobs, we should wait on it instead, but we
                        // would also want to receive shared jobs still?
                        // Spin? probably just wastes CPU time.
                        // self.context.shared_job.wait(&mut guard);
                        // if spin.spin() {
                        //     // wait for more shared jobs.
                        //     // self.context.shared_job.wait(&mut guard);
                        //     return;
                        // }
                        // Yield? same as spinning, really, so just exit and let the upstream use wait
                        // std::thread::yield_now();

                        self.heartbeat.latch.wait_and_reset();
                        // since we were sleeping, the shared job can't be populated,
                        // so resuming the inner loop is fine.
                    }
                }
            }
        }
        return;
    }

    pub fn wait_until_job<T>(&self, job: &Job<T>, latch: &CoreLatch) -> Option<JobResult<T>> {
        // we've already checked that the job was popped from the queue
        // check if shared job is our job
        if let Some(shared_job) = self.context.shared().jobs.remove(&self.index) {
            if core::ptr::eq(shared_job.as_ptr(), job as *const Job<T> as _) {
                // this is the job we are looking for, so we want to
                // short-circuit and call it inline
                return None;
            } else {
                // this isn't the job we are looking for, but we still need to
                // execute it
                self.execute(shared_job);
            }
        }

        // do the usual thing and wait for the job's latch
        if !latch.probe() {
            self.wait_until_latch_cold(latch);
        }

        Some(job.wait())
    }

    pub fn wait_until_latch<L>(&self, latch: &L)
    where
        L: AsCoreLatch,
    {
        let latch = latch.as_core_latch();
        if !latch.probe() {
            self.wait_until_latch_cold(latch)
        }
    }

    #[inline]
    fn wait_until_predicate<F>(&self, pred: F)
    where
        F: Fn() -> bool,
    {
        'outer: while !pred() {
            // take a shared job, if it exists
            if let Some(shared_job) = self.context.shared().jobs.remove(&self.index) {
                self.execute(shared_job);
            }

            // process local jobs before locking shared context
            while let Some(job) = self.pop_front() {
                unsafe {
                    job.as_ref().set_pending();
                }
                self.execute(job);
            }

            while !pred() {
                let mut guard = self.context.shared();
                let mut _spin = SpinWait::new();

                match guard.pop_job() {
                    Some(job) => {
                        drop(guard);
                        self.execute(job);

                        continue 'outer;
                    }
                    None => {
                        tracing::trace!("waiting for shared job, thread id: {:?}", self.index);

                        // TODO: wait on latch? if we have something that can
                        // signal being done, e.g. can be waited on instead of
                        // shared jobs, we should wait on it instead, but we
                        // would also want to receive shared jobs still?
                        // Spin? probably just wastes CPU time.
                        // self.context.shared_job.wait(&mut guard);
                        // if spin.spin() {
                        //     // wait for more shared jobs.
                        //     // self.context.shared_job.wait(&mut guard);
                        //     return;
                        // }
                        // Yield? same as spinning, really, so just exit and let the upstream use wait
                        // std::thread::yield_now();
                        return;
                    }
                }
            }
        }
        return;
    }
}
